# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_logr.ipynb.

# %% auto 0
__all__ = ['LogMessage', 'RecurrentLogMessage', 'EncoderLogMessage', 'DecoderLogMessage']

# %% ../nbs/02_logr.ipynb 6
import logging

# %% ../nbs/02_logr.ipynb 8
from typing import Mapping

# %% ../nbs/02_logr.ipynb 11
#| export

# %% ../nbs/02_logr.ipynb 13
#| export


# %% ../nbs/02_logr.ipynb 15
try: import torch, torch.nn as nn
except ImportError: ...

# %% ../nbs/02_logr.ipynb 17
#| export


# %% ../nbs/02_logr.ipynb 19
from atyp import IntQ, StrQ, SysExcInfoType
from chck import isnone, notnone

# %% ../nbs/02_logr.ipynb 21
#| export


# %% ../nbs/02_logr.ipynb 23
class LogMessage:
    '''
    A class for creating structured log messages.

    Attributes
    ----------
    name : str
        The name of the log message type.
    level : int
        The logging level.

    Methods
    -------
    __init__(msg: StrQ, *args, **kwargs):
        Initializes the log message.
    register_level():
        Registers a new logging level.
    append_extras(msg: StrQ, extras: dict):
        Appends extra information to the log message.
    __call__(logger, *args, **kwargs):
        Logs the message using the specified logger.

    Examples
    --------
    >>> log_msg = LogMessage("Example message")
    >>> log_msg(logger)
    # Logs "Example message" at the INFO level.
    '''
    name: str = None
    level: int = logging.INFO
    
    def __init__(self, msg: StrQ = '', *args, **kwargs):
        self.msg = msg if notnone(msg) else ''
        self.extras = kwargs
        type(self).register_level()

    @classmethod
    def register_level(cls):
        name = cls.name or cls.__name__
        level = cls.level or logging.INFO
        if isnone(name): name = logging.getLevelNamesMapping().get(name, None)
        if notnone(name): logging.addLevelName(level, name)

    @classmethod
    def append_extras(cls, msg: StrQ = '', extras: dict = dict()) -> str:
        if isnone(msg): msg = ''
        estr = ', '.join(f'{k}={v}' for k, v in extras.items())
        msg += f'\textras({estr})'
        return msg

    def __call__(
        self, logger: logging.Logger = None, *args, 
        exc_info: SysExcInfoType = None, stack_info: bool = False, stacklevel: int = 1, 
        extras: Mapping[str, object] = None, **kwargs
    ):
        extras = dict() if isnone(extras) else extras
        extras = {**getattr(self, 'extras', dict()), **extras, **kwargs}
        lvl = type(self).level
        msg = self.append_extras(self.msg, extras)
        if kwargs.pop('print', False): print(msg)
        if isnone(logger): return
        logger.log(lvl, msg, *args, exc_info=exc_info, stack_info=stack_info, stacklevel=stacklevel, extras=extras)

# %% ../nbs/02_logr.ipynb 24
class RecurrentLogMessage(LogMessage):
    '''
    A log message class specialized for recurrent neural network operations.

    Inherits from LogMessage and adds specific attributes for RNN logging.

    Methods
    -------
    __init__(msg: StrQ, xshape, hshape, cshape, seqlen, step, step_idx, *args, **kwargs):
        Initializes the recurrent log message.

    Examples
    --------
    >>> rnn_log_msg = RecurrentLogMessage("RNN step", xshape=(10, 5), hshape=(2, 10, 5))
    >>> rnn_log_msg(logger)
    # Logs a message with RNN-specific details.
    '''
    def __init__(
        self, msg: StrQ = '', xshape: IntQ = None, hshape: IntQ = None, cshape: IntQ = None, 
        seqlen: IntQ = None, step: StrQ = None, step_idx: IntQ = None, *args, **kwargs
    ):
        attrs = ('xshape', 'hshape', 'cshape', 'seqlen', 'step', 'step_idx')
        extras = dict(zip(attrs, (step, xshape, hshape, cshape, seqlen, step_idx)))
        super().__init__(msg, *args, **extras, **kwargs)
        self.extras = extras

# %% ../nbs/02_logr.ipynb 25
class EncoderLogMessage(RecurrentLogMessage):
    '''
    A log message class for logging encoder-specific information in an autoencoder.

    Inherits from RecurrentLogMessage and sets a specific name and level for encoder logging.
    '''
    name = 'Encoder'
    level = 22
    
class DecoderLogMessage(RecurrentLogMessage):
    '''
    A log message class for logging decoder-specific information in an autoencoder.

    Inherits from RecurrentLogMessage and sets a specific name and level for decoder logging.
    '''
    name = 'Decoder'
    level = 23
